/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"

//////////////////////////////////////////////////////////////////
//
// Reorganization Test #2 - Randomly delete and reinsert as many
// atomic parts per composite part as each one has, interleaving
// reorganizations of all of the composite parts, to test out the
// deletion/reallocation handling and performance characteristics
// of the system
//
//////////////////////////////////////////////////////////////////


extern char*	types[NumTypes];

extern ooTrans	xact;
extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg2
//
//////////////////////////////////////////////////////////////////


int reorg2(ooHandle(ooDBObj)& databaseH)
{

    // get access to the design library holding composite parts

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocUpdate);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }

    // for as many atomic parts as each composite part contains,
    // iterate through all the composite parts, deleting and then
    // recreating one of the atomic parts from its atomic parts graph
    // (and count the number of atomic parts processed in this way)

    int count = 0;
    ooItr(CompositePart) compI;
    for (int i = 0; i < NumAtomicPerComp; i++) {
        for (compI.scan(designLibH, oocUpdate); compI.next(); ) {
            if (debugMode) {
                printf("    In Reorg2, partId = %d:\n", compI->id);
            }
            count += compI->reorg2(designLibH);
        }
    }

    return count;
}



//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg2
//
//////////////////////////////////////////////////////////////////


int CompositePart::reorg2(ooHandle(ooContObj)& designLibH)
{
    if (debugMode) {
        printf("CompositePart::Reorg2(id = %d)\n", id);
    }

    // first select and locate a victim atomic part

    int lowId, victimId, id;
    ooKey key;
    key.type = oocInt32;
    key.offset = ooGetMemberOffset(AtomicPart, id);
    key.size   = ooGetMemberSize(AtomicPart, id);
    key.value  = &id;
    ooHandle(CompositePart) thisH = ooThis();
    lowId = rootPart()->id;
    victimId = lowId + (int) (random() % NumAtomicPerComp);
    id = victimId;
    ooHandle(AtomicPart) victimH;
    victimH.lookupObj(designLibH, key);

    // next record the affected to and from connections

    PartIdList toList;
    PartIdList fromList;
    ooItr(Connection) connI;

    for (victimH->to(connI); connI.next(); ) {
	toList.insert(connI->to()->id);
    }
    for (victimH->from(connI); connI.next(); ) {
	if (connI->from()->id != victimH->id) {
	    fromList.insert(connI->from()->id);
        }
    }

    // now do the actual part deletion

    if (debugMode) {
        printf("\tReorg2: Deleting atomic part %d\n", victimH->id);
    }
    ooDelete(victimH);

    // time to (approximately) recreate the deleted part

    int32 partDate;
    int typeNo, x, y, length;
    typeNo = (int) (random() % NumTypes);
    partDate = MinAtomicDate + 
	       (int) (random() % (MaxAtomicDate-MinAtomicDate+1));
    x = (int) (random() % XYRange);
    y = (int) (random() % XYRange);
    victimH = ooNewKey(AtomicPart, (victimId), thisH, key);
    victimH->set_partOf(thisH);
    if (victimId == lowId) {
       set_rootPart(victimH);
    }
    if (debugMode) {
        printf("\tReorg2: Recreated atomic part %d\n", victimH->id);
    }

    // finally, connect the part back up to the other parts again

    ooHandle(AtomicPart) partH;
    ooHandle(Connection) connH;

    while (!toList.empty()) {
	id = toList.remove();
	typeNo = (int) (random() % NumTypes);
	length = (int) (random() % XYRange);
        partH.lookupObj(designLibH, key);
	connH = new(thisH) Connection(victimH, partH);
    }

    while (!fromList.empty()) {
	id = fromList.remove();
	typeNo = (int) (random() % NumTypes);
	length = (int) (random() % XYRange);
        partH.lookupObj(designLibH, key);
	connH = new(thisH) Connection(partH, victimH);
    }

    return 1;
}

