/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"

/////////////////////////////////////////////////////////////////
//
// Traversal #7.  Pick a random atomic part; traverse from the 
// part to its containing composite part P.  From P, traverse
// backwards up to all base assemblies referencing P; for each
// such base assembly, traverse upward through the assembly
// hierarchy until the root is reached.  This traversal coupled
// with traversal #1 ensures
// that all pointers are bi-directional.
//
/////////////////////////////////////////////////////////////////

extern ooTrans	xact;
extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Routine to initiate Traversal #7
//
//////////////////////////////////////////////////////////////////


int traverse7(ooHandle(ooDBObj)& databaseH)
{

    // get access to the design library containing atomic parts

    ooHandle(ooContObj) designLibH;
    designLibH.open(databaseH, "DesignLibrary", oocRead);
    if (designLibH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to access DesignLibrary\n");
	xact.abort();
        exit(1);
    }


    // prepare to select a random atomic part via the id index

    ooKey key;
    int partId;
    key.type = oocInt32;
    key.offset = ooGetMemberOffset(AtomicPart, id);
    key.size   = ooGetMemberSize(AtomicPart, id);
    key.value  = &partId;
    ooHandle(AtomicPart) atomH;

    // now randomly select a part and traverse upwards from there

    // set random seed so "hot" runs are truly hot
    srandom(1);

    partId = (int) (random() % TotalAtomicParts) + 1;
    atomH.lookupObj(designLibH, key);
    if (atomH.isValid() == oocFalse) {
        fprintf(stderr, "ERROR: Unable to find atomic part %d\n", partId);
   	xact.abort();
        exit(1);
    }

    if (debugMode) {
        printf("  In Traverse7, starting at atomic part %d\n", partId);
	printf("  In Traverse7, traversing to composite part %d\n",
		atomH->partOf()->id);

    }

    return atomH->partOf()->traverse7();
}



//////////////////////////////////////////////////////////////////
//
// CompositePart Method for Traversal #7
//
//////////////////////////////////////////////////////////////////


int CompositePart::traverse7()
{
    if (debugMode) {
        printf("\tCompositePart::traverse7(id = %d)\n", id);
    }

    // search up the design hierarchy (along the private path)

    PartIdSet visitedBaseIds;
    PartIdSet visitedComplexIds;
    ooItr(BaseAssembly) baseI;
    int count = 0;
    for (usedInPriv(baseI); baseI.next(); ) {
        if (debugMode) {
            printf("\tConsidering base assembly %d\n", baseI->id);
        }
        if (!visitedBaseIds.contains(baseI->id)) {
            visitedBaseIds.insert(baseI->id);
	    count += baseI->traverse7(visitedComplexIds);
	}
    }
    return count;
}


//////////////////////////////////////////////////////////////////
//
// BaseAssembly Method for Traversal #7
//
//////////////////////////////////////////////////////////////////


int BaseAssembly::traverse7(PartIdSet& visitedComplexIds)
{
    if (debugMode) {
        printf("\t\tBaseAssembly::traverse7(id = %d)\n", id);
    }

    // visit this assembly and move on up the design hierarchy
    int count = 1;

    DoNothing(); // invoke DoNothing method
    if (!visitedComplexIds.contains(superAssembly()->id)) {
        count += superAssembly()->traverse7(visitedComplexIds);
    }
    return count;
}


//////////////////////////////////////////////////////////////////
//
// ComplexAssembly Method for Traversal #7
//
//////////////////////////////////////////////////////////////////


int ComplexAssembly::traverse7(PartIdSet& visitedComplexIds)
{
    if (debugMode) {
        printf("\t\t\tComplexAssembly::traverse7(id = %d)\n", id);
    }

    // process this assembly and move on up the design hierarchy

    int count = 1;

    visitedComplexIds.insert(id);
    DoNothing(); // invoke  DoNothing Method
    if (!superAssembly().isNull()) {
	if (!visitedComplexIds.contains(superAssembly()->id)) {
            count += superAssembly()->traverse7(visitedComplexIds);
        }
    }
    return count;
}

